%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Definitions of builtins of module IOExts:
%

:- use_module(prim_ports). % for IOExts.choiceStreamOrMsgs

:- dynamic globalAssoc/2.

?- block 'prim_setAssoc'(?,?,?,?,-,?).
'prim_setAssoc'(RKey,RVal,_,'$io'('()'),E0,E) :-
        derefAll(RKey,Key), derefAll(RVal,Val),
        string2Atom(Key,KeyA),
	(retract(globalAssoc(KeyA,_)) -> true ; true),
	assertz(globalAssoc(KeyA,Val)),
	!, E0=E.

?- block 'prim_getAssoc'(?,?,?,-,?).
'prim_getAssoc'(RKey,_,'$io'(R),E0,E) :-
        derefAll(RKey,Key),
        string2Atom(Key,KeyA),
	(globalAssoc(KeyA,Val) -> R='Just'(Val) ; R='Nothing'),
	!, E0=E.

% shell command execution:
?- block prim_execCmd(?,?,?,-,?).
prim_execCmd(RCmdString,_,'$io'('(,,)'(StdIn,StdOut,StdErr)),E0,E) :-
        derefAll(RCmdString,CmdString),
	string2Atom(CmdString,Cmd),
	exec(Cmd,[pipe(StdIn),pipe(StdOut),pipe(StdErr)],_), E0=E.


% open a connection to a Unix socket:
?- block prim_connectToSocket(?,?,?,?,-,?).
prim_connectToSocket(RSHst,RSNr,_,'$io'(SocketStream),E0,E) :-
	derefAll(RSHst,SHst), deref(RSNr,SNr),
        string2Atom(SHst,Host), !,
        socket('AF_INET',Socket),
        socket_connect(Socket,'AF_INET'(Host,SNr),SocketStream), E0=E.


% choice on a stream and an external port message stream:
?- block prim_choiceStreamOrMsgs(?,-,?,?,?,?),
         prim_choiceStreamOrMsgs(?,?,?,?,-,?).
prim_choiceStreamOrMsgs(Handle,share(M),World,Result,E0,E) :-
        !,
	get_mutable(V,M),
	(V='$eval'(R) % external message stream has been already evaluated
	 -> E0=E, Result='$io'('Right'(R))
	  ; prim_choiceStreamOrMsgs(Handle,V,World,CResult,E0,E),
	    (CResult='$io'('Left'(_))
  	     -> Result=CResult
	      ; CResult='$io'('Right'(S)),
	        (compileWithSharing(variable) -> propagateShare(S,TResult) ; S=TResult),
	        Result='$io'('Right'(TResult)),
	        update_mutable('$eval'(TResult),M))).
prim_choiceStreamOrMsgs(_,[M|Ms],_,'$io'('Right'([M|Ms])),E0,E) :-
	!, E0=E. % stream already evaluated
prim_choiceStreamOrMsgs(Handle,[],_,'$io'('Left'(Handle)),E0,E) :-
	% message stream is empty, so anything must be received from handle stream.
	!, E0=E.
prim_choiceStreamOrMsgs(Handle,'Ports.basicServerLoop'(Port),World,
                                 Result,E0,E) :-
        Port='Ports.internalPort'(_,_,PNr,_),
	checkIncomingPortStreams(PNr,PortStream),
	!,
	readStreamUntilZero(PortStream,MsgString),
	(parse_received_message(PortStream,MsgString,Msg)
	 -> ifTracePort((write(user_error,'TRACEPORTS: Received message: '),
		         write(user_error,Msg), nl(user_error))),
	    Result='$io'('Right'([Msg|'Ports.basicServerLoop'(Port)])), E0=E
	  ; write(user_error,'ERROR: Illegal message received (ignored): '),
	    putChars(user_error,MsgString), nl(user_error),
	    prim_choiceStreamOrMsgs(Handle,'Ports.basicServerLoop'(Port),World,
	                                   Result,E0,E)),
	!.
prim_choiceStreamOrMsgs(HandleExp,PortStream,World,Result,E0,E):-
	hnf(HandleExp,Handle,E0,E1),
	prim_choiceStreamOrMsgs_select(Handle,PortStream,World,Result,E1,E).

?- block prim_choiceStreamOrMsgs_select(?,?,?,?,-,?).
prim_choiceStreamOrMsgs_select(Handle,'Ports.basicServerLoop'(Port),World,
			       Result,E0,E):-
        Port='Ports.internalPort'(_,_,PNr,Socket),
	socket_select([Socket],PortStreams,Clients,off,[Handle],_SelStreams),
	(PortStreams=[PortStream]
	 -> Clients=[Client],
	    ifTracePort((write(user_error,'TRACEPORTS: Connection to client: '),
		         write(user_error,Client), nl(user_error))),
	    (readPortMessage(PNr,PortStream,MsgString)
	     -> (parse_received_message(PortStream,MsgString,Msg)
	         -> ifTracePort((write(user_error,'TRACEPORTS: Received message: '),
		                 write(user_error,Msg), nl(user_error))),
		    Result='$io'('Right'([Msg|'Ports.basicServerLoop'(Port)])), E0=E
	          ; write(user_error,'ERROR: Illegal message received (ignored): '),
		    putChars(user_error,MsgString), nl(user_error),
		    prim_choiceStreamOrMsgs(Handle,
                                     'Ports.basicServerLoop'(Port),World,Result,E0,E))
	      ; % try reading next message:
		prim_choiceStreamOrMsgs(Handle,
		                     'Ports.basicServerLoop'(Port),World,Result,E0,E))
	  ; % SelStreams must be Handle:
	    Result='$io'('Left'(Handle)), E0=E),
	!.




%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% primitives implementing IORefs:

% New IORefs are represented as mutable values. The "share" constructor
% is put around to be conform with the remaining implementation where
% all mutables are "marked" by this constructor.
?- block 'prim_newIORef'(?,?,?,-,?).
'prim_newIORef'(Val,_,'$io'('IOExts.IORef'(share(MutVal))),E0,E) :-
        var(Val), !,
	create_mutable('$eval'(Val),MutVal), E0=E.
'prim_newIORef'(Val,_,'$io'('IOExts.IORef'(share(MutVal))),E0,E) :-
	create_mutable(Val,MutVal), E0=E.

% When an IORef is read and its value is not evaluated, the current value
% is wrapped into a new mutable in order to implement sharing
% of evaluations of IORefs. The current IORef is updated so that
% it refers to the new mutable (without this indirection, there is
% a risk of creating cyclic structures when the IORef itself is updated).
?- block prim_readIORef(?,?,?,-,?).
prim_readIORef(IORef,_,R,E0,E) :-
        hnf(IORef,HIORef,E0,E1),
	prim_readIORef_exec(HIORef,R,E1,E).

?- block prim_readIORef_exec(-,?,?,?), prim_readIORef_exec(?,?,-,?).
prim_readIORef_exec('IOExts.IORef'(share(MutVal)),'$io'(V),E0,E) :-
        get_mutable(Val,MutVal),
	(Val='$eval'(V) -> true
	  ; create_mutable(Val,MutV),
	    update_mutable(share(MutV),MutVal),
	    V=share(MutV)),
        E0=E.

% Assign a new value to an IORef:
?- block prim_writeIORef(?,?,?,?,-,?).
prim_writeIORef(IORef,Val,_,R,E0,E) :-
        hnf(IORef,HIORef,E0,E1),
	prim_writeIORef_exec(HIORef,Val,R,E1,E).

?- block prim_writeIORef_exec(-,?,?,?,?), prim_writeIORef_exec(?,?,?,-,?).
prim_writeIORef_exec('IOExts.IORef'(share(MutVal)),Val,'$io'('()'),E0,E) :-
        var(Val), !,
	update_mutable('$eval'(Val),MutVal), E0=E.
prim_writeIORef_exec('IOExts.IORef'(share(MutVal)),Val,'$io'('()'),E0,E) :-
	update_mutable(Val,MutVal), E0=E.

