%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Definitions of builtins of module System:
%

:- use_module(library(system)).

?- block 'prim_getCPUTime'(?,?,-,?).
'prim_getCPUTime'(_,'$io'(MS),E0,E) :-
	statistics(runtime,[MS,_]), E0=E.

?- block 'prim_getElapsedTime'(?,?,-,?).
'prim_getElapsedTime'(_,'$io'(MS),E0,E) :-
	statistics(walltime,[MS,_]), E0=E.

?- block 'prim_getDate'(?,?,-,?).
'prim_getDate'(_,'$io'('System.Date'(Y,M,D,H,Min,S)),E0,E) :-
        datime(datime(Y,M,D,H,Min,S)), E0=E.

?- block 'prim_getArgs'(?,?,-,?).
'prim_getArgs'(_,'$io'(StringArgs),E0,E) :-
        (rtargs(Args) -> true ; prolog_flag(argv,Args,Args)),
        map2(atom2String,Args,StringArgs), E0=E.

?- block 'prim_getEnviron'(?,?,?,-,?).
'prim_getEnviron'(RVar,_,'$io'(Value),E0,E) :- derefAll(RVar,Var),
	string2Atom(Var,AtomVar),
	(environ(AtomVar,AtomValue) -> atom2String(AtomValue,Value)
	                             ; Value = []), % empty string if undefined
	E0=E.

?- block 'prim_getHostname'(?,?,-,?).
'prim_getHostname'(_,'$io'(String),E0,E) :-
        host_name(Name),
        atom2String(Name,String), E0=E.

?- block 'prim_getPID'(?,?,-,?).
'prim_getPID'(_,'$io'(Pid),E0,E) :- pid(Pid), E0=E.

?- block 'prim_getProgName'(?,?,-,?).
'prim_getProgName'(_,'$io'(String),E0,E) :-
        currentModule(Name),
        atom2String(Name,String), E0=E.

?- block 'prim_system'(?,?,?,-,?).
'prim_system'(RS,_,'$io'(Status),E0,E) :- derefAll(RS,S),
	string2Atom(S,Cmd),
	system(Cmd,Status), E0=E.

?- block 'prim_sleep'(?,?,?,-,?).
'prim_sleep'(RS,_,'$io'('()'),E0,E) :- deref(RS,S), sleep(S), E0=E.

% evaluate the argument to normal form and print the evaluation time:
?- block 'prim_evalTime'(?,?,-,?).
'prim_evalTime'(T,NF,E0,E) :-
	statistics(runtime,[RTime1,_]),
	statistics(walltime,[ETime1,_]),
	statistics(garbage_collection,[GC1,_,_]),
	nf(T,NF,E0,E1),
	statistics(runtime,[RTime2,_]),
	statistics(walltime,[ETime2,_]),
	statistics(garbage_collection,[GC2,_,_]),
	writeErr('EVALTIME: Runtime: '),
	RTime is RTime2-RTime1, writeErr(RTime),
	writeErr(' msec.'), nlErr,
	writeErr('EVALTIME: Elapsed time: '),
	ETime is ETime2-ETime1, writeErr(ETime),
	writeErr(' msec.'), nlErr,
	writeErr('EVALTIME: Number of garbage collections: '),
	GC is GC2-GC1, writeErr(GC), nlErr,
	prolog_flag(gc,_,on), E1=E.

% evaluate the argument to normal form (without garbage collection)
% and print the evaluation time and memory usage:
?- block 'prim_evalSpace'(?,?,-,?).
'prim_evalSpace'(T,NF,E0,E) :-
	garbage_collect,
	prolog_flag(gc,_,off),
	statistics(runtime,[RTime1,_]),
	statistics(walltime,[ETime1,_]),
	statistics(garbage_collection,[GC1,_,_]),
	statistics(global_stack,[GSU1,_]),
	statistics(local_stack,[LSU1,_]),
	nf(T,NF,E0,E1),
	statistics(runtime,[RTime2,_]),
	statistics(walltime,[ETime2,_]),
	statistics(garbage_collection,[GC2,_,_]),
	statistics(global_stack,[GSU2,_]),
	statistics(local_stack,[LSU2,_]),
	writeErr('EVALTIME: Runtime: '),
	RTime is RTime2-RTime1, writeErr(RTime),
	writeErr(' msec.'), nlErr,
	writeErr('EVALTIME: Elapsed time: '),
	ETime is ETime2-ETime1, writeErr(ETime),
	writeErr(' msec.'), nlErr,
	writeErr('EVALTIME: Number of garbage collections: '),
	GC is GC2-GC1, writeErr(GC), nlErr,
	writeErr('EVALTIME: Global stack usage: '),
	GS is GSU2-GSU1, writeErr(GS), nlErr,
	writeErr('EVALTIME: Local stack usage: '),
	LS is LSU2-LSU1, writeErr(LS), nlErr,
	prolog_flag(gc,_,on), E1=E.
